"""
This code provides a channel which can insert bit errors into messages
CrySyS Lab
2019
"""

import socket
import random
import logging
import sys
import message_format
import server_address

# Default parameters
BIT_ERROR_RATE_MESSAGE = 0.00
BIT_ERROR_RATE_ECC = 0.00


def add_client(client, c_list):
    """Add a new client to the current client list"""

    if client not in c_list:
        c_list.append(client)
    else:
        logging.warning("Client already added, not adding again")
    logging.info("Client list: " + str(c_list))


def remove_client(client, c_list):
    """Remove a client from the current client list"""

    try:
        c_list.remove(client)
    except ValueError:
        logging.warning("Client not connected yet, cannot disconnect")
    logging.info("Client list:" + str(c_list))


def create_random_mask(bit_error_rate):
    """Creates a random mask with the given error rate"""

    mask = ''
    # For each bit
    for _ in range(0, 8):
        if random.random() < bit_error_rate:
            mask = mask + '1'
        else:
            mask = mask + '0'
    return int(mask, 2)


def do_transfer(dat, sender, c_list, sock):
    """Receive a message -> Introduce some errors -> Broadcast it to the others"""

    logging.info("Message without error: " + str(dat))

    length_message = dat[message_format.LENGTH_MESSAGE_TYPE]
    logging.debug("Length of message: " + str(length_message))
    type_ecc = dat[message_format.LENGTH_MESSAGE_TYPE + message_format.LENGTH_MESSAGE_LENGTH + length_message]
    logging.debug("Type of ecc: " + str(type_ecc))
    length_ecc = dat[message_format.LENGTH_MESSAGE_TYPE + message_format.LENGTH_MESSAGE_LENGTH + length_message +
                     message_format.LENGTH_ECC_TYPE]
    logging.debug("Length of ecc " + str(length_ecc))

    # add message type and length to the buffer
    modified_dat = dat[0:message_format.LENGTH_MESSAGE_TYPE + message_format.LENGTH_MESSAGE_LENGTH]

    # introduce error to the message
    for i in range(message_format.LENGTH_MESSAGE_TYPE + message_format.LENGTH_MESSAGE_LENGTH,
                   message_format.LENGTH_MESSAGE_TYPE + message_format.LENGTH_MESSAGE_LENGTH + length_message):
        mask = create_random_mask(BIT_ERROR_RATE_MESSAGE)
        # XOR the error to the data
        erroneous_value = dat[i] ^ mask
        # add modified message byte to the buffer
        modified_dat = modified_dat + erroneous_value.to_bytes(1, byteorder="big")

    # add ecc type and length to the buffer
    ecc_t_and_l = dat[message_format.LENGTH_MESSAGE_TYPE + message_format.LENGTH_MESSAGE_LENGTH + length_message:
                      message_format.LENGTH_MESSAGE_TYPE + message_format.LENGTH_MESSAGE_LENGTH +
                      length_message + message_format.LENGTH_ECC_TYPE + message_format.LENGTH_ECC_LENGTH]
    modified_dat = modified_dat + ecc_t_and_l

    # introduce error to the ecc
    for i in range(message_format.LENGTH_MESSAGE_TYPE + message_format.LENGTH_MESSAGE_LENGTH + length_message +
                   message_format.LENGTH_ECC_TYPE + message_format.LENGTH_ECC_LENGTH,
                   message_format.LENGTH_MESSAGE_TYPE + message_format.LENGTH_MESSAGE_LENGTH + length_message +
                   message_format.LENGTH_ECC_TYPE + message_format.LENGTH_ECC_LENGTH + length_ecc):
        mask = create_random_mask(BIT_ERROR_RATE_ECC)
        # XOR the error to the data
        erroneous_value = dat[i] ^ mask
        # add modified ecc to the buffer
        modified_dat = modified_dat + erroneous_value.to_bytes(1, byteorder="big")

    logging.info("Message with error:   " + str(modified_dat))
    for client in c_list:
        if client != sender:
            logging.info("Sending to client: " + str(client))
            sock.sendto(modified_dat, client)


def main():
    """Main functionality"""

    global BIT_ERROR_RATE_MESSAGE
    global BIT_ERROR_RATE_ECC

    # Configure logging
    logging.basicConfig(format='%(asctime)s - %(message)s', level=logging.DEBUG)

    # Set the channel parameters
    if len(sys.argv) > 1:
        BIT_ERROR_RATE_MESSAGE = float(sys.argv[1])
        BIT_ERROR_RATE_ECC = float(sys.argv[1])

    logging.warning("Parameters of the channel: BER message: " + str(BIT_ERROR_RATE_MESSAGE) +
                    "; BER ecc: " + str(BIT_ERROR_RATE_ECC))

    # Create socket and bind to it
    sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)  # Internet, UDP
    sock.bind((server_address.UDP_IP, server_address.UDP_PORT))
    logging.warning("Connected to socket successfully, ready to serve")

    # We have just started, we have no clients yet
    client_list = []

    while True:
        data = bytearray()
        address = ()
        try:
            data, address = sock.recvfrom(server_address.UDP_MAX_BUF_SIZE)
        except ConnectionResetError:
            logging.warning("We have lost one of our clients without proper disconnect message")

        logging.debug("-----")
        logging.debug("Received message: " + str(data) + " from: " + str(address))
        received_type = data[0]
        logging.debug("Received type: " + str(received_type))

        if received_type == message_format.TYPE_CONNECT:
            logging.info("Received connect request")
            add_client(address, client_list)

        elif received_type == message_format.TYPE_DISCONNECT:
            logging.info("Received disconnect request")
            remove_client(address, client_list)

        elif received_type == message_format.TYPE_TRANSFER:
            logging.info("Received transfer request")
            do_transfer(data, address, client_list, sock)

        else:
            logging.warning("Received unknown request")


main()
